#!/usr/bin/perl -w

# Program Luhn_test.pl

# (c) Gary C. Kessler, 2014 [gck@garykessler.net]

# Program to test a number to determine if:
#  A) It is valid per Luhn's algorithm.
#  B) It appears to be a valid credit card number

# ********************************************************
# MAIN PROGRAM BLOCK
# ********************************************************

# Initialize global variables

$version = "1.4";
$build_date = "20 July 2014";

# The following array of information is largely based upon
# http://en.wikipedia.org/wiki/Bank_card_number

# This list is ordered in descending order of prefix length, to ensure best match.
# NOTE: Need "" around Maestro numbers or else they are interpreted as octal values!

@CardTypes =
  (
  { Name => "Visa Electron",          Prefix => 417500,   NumLen => 16 },
  { Name => "Maestro",                Prefix => "0604",   NumLen => 12 },
  { Name => "Maestro",                Prefix => "0604",   NumLen => 13 },
  { Name => "Maestro",                Prefix => "0604",   NumLen => 14 },
  { Name => "Maestro",                Prefix => "0604",   NumLen => 15 },
  { Name => "Maestro",                Prefix => "0604",   NumLen => 16 },
  { Name => "Maestro",                Prefix => "0604",   NumLen => 17 },
  { Name => "Maestro",                Prefix => "0604",   NumLen => 18 },
  { Name => "Maestro",                Prefix => "0604",   NumLen => 19 },
  { Name => "Visa Electron",          Prefix => 4026,     NumLen => 16 },
  { Name => "Visa Electron",          Prefix => 4405,     NumLen => 16 },
  { Name => "Visa Electron",          Prefix => 4508,     NumLen => 16 },
  { Name => "Visa Electron",          Prefix => 4844,     NumLen => 16 },
  { Name => "Visa Electron",          Prefix => 4913,     NumLen => 16 },
  { Name => "Visa Electron",          Prefix => 4917,     NumLen => 16 },
  { Name => "Maestro",                Prefix => 5018,     NumLen => 12 },
  { Name => "Maestro",                Prefix => 5018,     NumLen => 13 },
  { Name => "Maestro",                Prefix => 5018,     NumLen => 14 },
  { Name => "Maestro",                Prefix => 5018,     NumLen => 15 },
  { Name => "Maestro",                Prefix => 5018,     NumLen => 16 },
  { Name => "Maestro",                Prefix => 5018,     NumLen => 17 },
  { Name => "Maestro",                Prefix => 5018,     NumLen => 18 },
  { Name => "Maestro",                Prefix => 5018,     NumLen => 19 },
  { Name => "Dankort",                Prefix => 5019,     NumLen => 16 },
  { Name => "Maestro",                Prefix => 5020,     NumLen => 12 },
  { Name => "Maestro",                Prefix => 5020,     NumLen => 13 },
  { Name => "Maestro",                Prefix => 5020,     NumLen => 14 },
  { Name => "Maestro",                Prefix => 5020,     NumLen => 15 },
  { Name => "Maestro",                Prefix => 5020,     NumLen => 16 },
  { Name => "Maestro",                Prefix => 5020,     NumLen => 17 },
  { Name => "Maestro",                Prefix => 5020,     NumLen => 18 },
  { Name => "Maestro",                Prefix => 5020,     NumLen => 19 },
  { Name => "Maestro",                Prefix => 5038,     NumLen => 12 },
  { Name => "Maestro",                Prefix => 5038,     NumLen => 13 },
  { Name => "Maestro",                Prefix => 5038,     NumLen => 14 },
  { Name => "Maestro",                Prefix => 5038,     NumLen => 15 },
  { Name => "Maestro",                Prefix => 5038,     NumLen => 16 },
  { Name => "Maestro",                Prefix => 5038,     NumLen => 17 },
  { Name => "Maestro",                Prefix => 5038,     NumLen => 18 },
  { Name => "Maestro",                Prefix => 5038,     NumLen => 19 },
  { Name => "Maestro",                Prefix => 5612,     NumLen => 12 },
  { Name => "Maestro",                Prefix => 5612,     NumLen => 13 },
  { Name => "Maestro",                Prefix => 5612,     NumLen => 14 },
  { Name => "Maestro",                Prefix => 5612,     NumLen => 15 },
  { Name => "Maestro",                Prefix => 5612,     NumLen => 16 },
  { Name => "Maestro",                Prefix => 5612,     NumLen => 17 },
  { Name => "Maestro",                Prefix => 5612,     NumLen => 18 },
  { Name => "Maestro",                Prefix => 5612,     NumLen => 19 },
  { Name => "Maestro",                Prefix => 5893,     NumLen => 12 },
  { Name => "Maestro",                Prefix => 5893,     NumLen => 13 },
  { Name => "Maestro",                Prefix => 5893,     NumLen => 14 },
  { Name => "Maestro",                Prefix => 5893,     NumLen => 15 },
  { Name => "Maestro",                Prefix => 5893,     NumLen => 16 },
  { Name => "Maestro",                Prefix => 5893,     NumLen => 17 },
  { Name => "Maestro",                Prefix => 5893,     NumLen => 18 },
  { Name => "Maestro",                Prefix => 5893,     NumLen => 19 },
  { Name => "Discover",               Prefix => 6011,     NumLen => 16 },
  { Name => "Laser/Maestro",          Prefix => 6304,     NumLen => 16 },
  { Name => "Laser/Maestro",          Prefix => 6304,     NumLen => 17 },
  { Name => "Laser/Maestro",          Prefix => 6304,     NumLen => 18 },
  { Name => "Laser/Maestro",          Prefix => 6304,     NumLen => 19 },
  { Name => "Maestro",                Prefix => 6390,     NumLen => 12 },
  { Name => "Maestro",                Prefix => 6390,     NumLen => 13 },
  { Name => "Maestro",                Prefix => 6390,     NumLen => 14 },
  { Name => "Maestro",                Prefix => 6390,     NumLen => 15 },
  { Name => "Maestro",                Prefix => 6390,     NumLen => 16 },
  { Name => "Maestro",                Prefix => 6390,     NumLen => 17 },
  { Name => "Maestro",                Prefix => 6390,     NumLen => 18 },
  { Name => "Maestro",                Prefix => 6390,     NumLen => 19 },
  { Name => "Laser",                  Prefix => 6706,     NumLen => 16 },
  { Name => "Laser",                  Prefix => 6706,     NumLen => 17 },
  { Name => "Laser",                  Prefix => 6706,     NumLen => 18 },
  { Name => "Laser",                  Prefix => 6706,     NumLen => 19 },
  { Name => "Laser",                  Prefix => 6709,     NumLen => 16 },
  { Name => "Laser",                  Prefix => 6709,     NumLen => 17 },
  { Name => "Laser",                  Prefix => 6709,     NumLen => 18 },
  { Name => "Laser",                  Prefix => 6709,     NumLen => 19 },
  { Name => "Maestro",                Prefix => 6759,     NumLen => 12 },
  { Name => "Maestro",                Prefix => 6759,     NumLen => 13 },
  { Name => "Maestro",                Prefix => 6759,     NumLen => 14 },
  { Name => "Maestro",                Prefix => 6759,     NumLen => 15 },
  { Name => "Maestro",                Prefix => 6759,     NumLen => 16 },
  { Name => "Maestro",                Prefix => 6759,     NumLen => 17 },
  { Name => "Maestro",                Prefix => 6759,     NumLen => 18 },
  { Name => "Maestro",                Prefix => 6759,     NumLen => 19 },
  { Name => "Laser",                  Prefix => 6771,     NumLen => 16 },
  { Name => "Laser",                  Prefix => 6771,     NumLen => 17 },
  { Name => "Laser",                  Prefix => 6771,     NumLen => 18 },
  { Name => "Laser",                  Prefix => 6771,     NumLen => 19 },
  { Name => "Diners Club Int.",       Prefix => 300,      NumLen => 14 },
  { Name => "Diners Club Int.",       Prefix => 301,      NumLen => 14 },
  { Name => "Diners Club Int.",       Prefix => 302,      NumLen => 14 },
  { Name => "Diners Club Int.",       Prefix => 303,      NumLen => 14 },
  { Name => "Diners Club Int.",       Prefix => 304,      NumLen => 14 },
  { Name => "Diners Club Int.",       Prefix => 305,      NumLen => 14 },
  { Name => "Diners Club Int.",       Prefix => 309,      NumLen => 14 },
  { Name => "Discover",               Prefix => 622,      NumLen => 16 },
  { Name => "InterPayment",           Prefix => 636,      NumLen => 16 },
  { Name => "InterPayment",           Prefix => 636,      NumLen => 17 },
  { Name => "InterPayment",           Prefix => 636,      NumLen => 18 },
  { Name => "InterPayment",           Prefix => 636,      NumLen => 19 },
  { Name => "InstaPayment",           Prefix => 637,      NumLen => 16 },
  { Name => "InstaPayment",           Prefix => 638,      NumLen => 16 },
  { Name => "InstaPayment",           Prefix => 639,      NumLen => 16 },
  { Name => "Discover",               Prefix => 644,      NumLen => 16 },
  { Name => "Discover",               Prefix => 645,      NumLen => 16 },
  { Name => "Discover",               Prefix => 646,      NumLen => 16 },
  { Name => "Discover",               Prefix => 647,      NumLen => 16 },
  { Name => "Discover",               Prefix => 648,      NumLen => 16 },
  { Name => "Discover",               Prefix => 649,      NumLen => 16 },
  { Name => "Maestro",                Prefix => 676,      NumLen => 12 },
  { Name => "Maestro",                Prefix => 676,      NumLen => 13 },
  { Name => "Maestro",                Prefix => 676,      NumLen => 14 },
  { Name => "Maestro",                Prefix => 676,      NumLen => 15 },
  { Name => "Maestro",                Prefix => 676,      NumLen => 16 },
  { Name => "Maestro",                Prefix => 676,      NumLen => 17 },
  { Name => "Maestro",                Prefix => 676,      NumLen => 18 },
  { Name => "Maestro",                Prefix => 676,      NumLen => 19 },
  { Name => "American Express",       Prefix => 34,       NumLen => 15 },
  { Name => "JCB",                    Prefix => 35,       NumLen => 16 },
  { Name => "Diners Club Int.",       Prefix => 36,       NumLen => 14 },
  { Name => "American Express",       Prefix => 37,       NumLen => 15 },
  { Name => "Diners Club Int.",       Prefix => 38,       NumLen => 14 },
  { Name => "Diners Club Int.",       Prefix => 39,       NumLen => 14 },
  { Name => "MasterCard",             Prefix => 50,       NumLen => 16 },
  { Name => "MasterCard",             Prefix => 51,       NumLen => 16 },
  { Name => "MasterCard",             Prefix => 52,       NumLen => 16 },
  { Name => "MasterCard",             Prefix => 53,       NumLen => 16 },
  { Name => "MasterCard/Diners Club", Prefix => 54,       NumLen => 16 },
  { Name => "MasterCard/Diners Club", Prefix => 55,       NumLen => 16 },
  { Name => "China UnionPay",         Prefix => 62,       NumLen => 16 },
  { Name => "China UnionPay",         Prefix => 62,       NumLen => 17 },
  { Name => "China UnionPay",         Prefix => 62,       NumLen => 18 },
  { Name => "China UnionPay",         Prefix => 62,       NumLen => 19 },
  { Name => "Discover",               Prefix => 65,       NumLen => 16 },
  { Name => "China UnionPay",         Prefix => 88,       NumLen => 16 },
  { Name => "China UnionPay",         Prefix => 88,       NumLen => 17 },
  { Name => "China UnionPay",         Prefix => 88,       NumLen => 18 },
  { Name => "China UnionPay",         Prefix => 88,       NumLen => 19 },
  { Name => "Visa",                   Prefix => 4,        NumLen => 13 },
  { Name => "Visa",                   Prefix => 4,        NumLen => 16 },
  );

# action will indicate print table, or test single entry or file entries

$action = "";

#
# Begin program activity!!
#

print "\nLuhn algorithm test/Credit card identifier V$version - Gary C. Kessler ($build_date)\n\n";


#
# Parse command line. If no command line switches, ask for card number
# If -n, skip request for number.
# If -i, read from file
# If -p, print card info
# If -h, print help file
#

# Parse command for file name and request missing information.

if (parsecommandline ())
  { exit 1; };

# At this point, $action can have one of four values: null, "print", "file", or "single"

ACTIONS:
  {

# No switches provided. Ask for a card number, then process...

  $action eq "" && do
    {
    print "Enter a card number: ";
    $card_num = <STDIN>;
    print "\n";
    chomp($card_num);
    main ();
    };

# Read test numbers from a file (-f switch)

  $action eq "file" && do
    {

# Open input ($source) file and stop program if error

    open (INFILE, $source) or die "Can't open input file \"$source\": $!";

    while (defined ($card_num = <INFILE>))
      {
      $card_num =~ s/\s+$//;
      main ();
      };

    print "\n";
    };

# Test a single credit card entry (-n switch)

  $action eq "single" && do
    {
    main ();
    };

# Display the credit card test template (-p switch)

  $action eq "print" && do
    {
    for $i ( 0 .. $#CardTypes )
      {
      print "$i is { ";
      for $role ( keys %{ $CardTypes[$i] } )
        { print "$role=$CardTypes[$i]{$role} "; }
      print "}\n";
      }
    print "\n";
    };

 };

print "Done!\n\n";


# ********************************************************
# *****           SUBROUTINE BLOCKS                  *****
# ********************************************************

# ********************************************************
# help_text
# ********************************************************

# Display the help file

sub help_text
{
print<< "EOT";
Program usage: Luhn_test {[-i input_file] | [-n card_number]}
               Luhn_test {[-h] | [-p]}

 where: -i is the input file name containing credit card numbers to
            test, one number per line
        -n is a single credit card number to test
        -h displays this help file
        -p displays the credit card template information (users may
             wish to pipe the output through more)

If no switch is provided, the program will ask for a single credit
card value.

EOT
return;
}

# ********************************************************
# main
# ********************************************************

# Main program activity occurs here. This is all in a subroutine so that the
# program can easily accomodate either a single input or input from a file

sub main
{
local ($i, $ok);

$card_len = length ($card_num);
print "Credit card number: $card_num Number length = ", $card_len, "\n";

if (validate_Luhn ($card_num))
    {
    print "Card has a valid Luhn algorithm value";

    $ok = 99;
    $i = 0;
    do
      {
      if (substr ($card_num,0,length($CardTypes[$i]{Prefix})) == $CardTypes[$i]{Prefix} && $card_len == $CardTypes[$i]{NumLen})
          { $ok = $i; }
        else
          { $i++; }
      }
    while ($ok == 99 && $i <= $#CardTypes);

    if ($ok == 99)
        { print "...\n...but has an unknown credit card type or incorrect card number length.\n\n"; }
      else
        { print "\nand appears to be a valid $CardTypes[$i]{Name} credit card number.\n\n"; }
     }
  else { print "Card does not have a valid Luhn algorithm value.\n\n"; }
}

# ********************************************************
# parsecommandline
# ********************************************************

# Parse command line for file name, if present. Query
# user for any missing information

# Return $state = 1 to indicate that the program should stop
# immediately (switch -h)

sub parsecommandline
{
my $state = 0;

# Parse command line switches ($ARGV array of length $#ARGV)

if ($#ARGV >= 0)
  { 
  for ($i = 0; $i <= $#ARGV; $i++)
    {
    PARMS:
      {
      $ARGV[$i] eq "-i" && do
         {
         $source = $ARGV[$i+1];
         $action = "file";
         return $state;
         };
      $ARGV[$i] eq "-p" && do
         {
         $action = "print";
         return $state;
         };
      $ARGV[$i] eq "-n" && do
         {
         $card_num = $ARGV[$i+1];
         $action = "single";
         return $state;
         };
      $ARGV[$i] eq "-h" && do
         {
         help_text();
         $state = 1;
         return $state;
         };

      do
         {
         print "Invalid parameter \"$ARGV[$i]\".\n\n";
         print "Usage: Luhn_test {[-i input_file] | [-n card_number]}\n";
         print "       Luhn_test {[-h] | [-p]}\n\n";
         $state = 1;
         return $state;
         };

      };
    };
  };

return $state;
}

# ********************************************************
# validate_cc
# ********************************************************

# Validate a number according to Luhn's algorithm. See
# http://en.wikipedia.org/wiki/Luhn_algorithm for details on the algorithm.

# This code snippet from
# http://rosettacode.org/wiki/Luhn_test_of_credit_card_numbers#Perl

sub validate_Luhn
{
my @rev = reverse split //,$_[0];
my ($sum1,$sum2,$i) = (0,0,0);
 
for (my $i=0;$i<@rev;$i+=2)
  {
  $sum1 += $rev[$i];
  last if $i == $#rev;
  $sum2 += 2*$rev[$i+1]%10 + int(2*$rev[$i+1]/10);
  }
return ($sum1+$sum2) % 10 == 0;
}


